/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mitaka.h"
#include  "mtk_mode.h"
#include  "mtk_planets.h"
#include  "mtk_objects.h"
#include  "mtk_object_position.h"
#include  "mtk_planet_position.h"
#include  "mtk_planet_orientation.h"
#include  "mtk_planet_phys_params.h"
#include  "mtk_satellite_data.h"
#include  "mtk_solar_system.h"
#include  "mtk_sun.h"
#include  "mtk_satellites.h"
#include  "mtk_minor_planets.h"
#include  "mtk_planetarium.h"
#include  "mtk_planet_eclipse.h"
#include  "mtk_label.h"
#include  "mtk_calc_alpha.h"
#include  "mtk_time.h"
#include  "mtk_stars.h"
#include  "mtk_place_name_earth_mars.h"
#include  "mtk_switches.h"
#include  "mtk_atmosphere.h"
#include  "mtk_font.h"
#include  "mtk_text.h"
#include  "mtk_gl_common.h"
#include  "mtk_draw_common.h"
#include  "mtk_planetary_ring.h"
#include  "mtk_dialog_box.h"
#include  "mtk_OpenGL.h"
#include  "../common/astro/astro_const.h"
#include  "../common/graphics/gl_triangle_mesh.h"
#include  "mtk_triangle_mesh.h"
#include  "mtk_colors.h"
#include  "mtk_eclipse.h"
#include  "mtk_target.h"




//-----------
//  Externs
//-----------
extern bool bDrawLowRes;	//****
extern TopoSphere	tpMoon;	// n`


vector2d	get_landing_pos();





//------------
//  Lighting
//------------
float AtmosphereIntensity;
float AtmosphereIntensityDef = 6.0;
float EarthSpecular;
float EarthSpecularDef = 6.0;
float EarthShininess;
float EarthShininessDef = 60;



//----------
//  Others
//----------
int EarthCityLightsIniDef = 0;
bool EarthCityLightsIni   = false;



//-------------
//  Constants
//-------------
const int PLANET_NUM = 9;
const char szPlanetTexFileName[PLANET_NUM][20] = {
	"mercury.jpg", "venus.jpg", "earth.jpg", "mars.jpg",
	"jupiter.jpg", "saturn.jpg", "uranus.jpg", "neptune.jpg", "pluto.jpg"
};


const int EARTH_DETAIL_TEXTURE_NUM = 3;
const char szEarthDetailTexFileName[EARTH_DETAIL_TEXTURE_NUM][30] = {
	"earth_japan.png",		// {
	"earth_hawaii.png",		// nC
	"earth_chile.png"		// `AALMA
};

float TopoZoomRate[5] = {
	1.0f, 2.0f, 5.0f, 10.0f, 20.0f
};




//-----------
//  Objects
//-----------

//--  TopoSphere  --
TopoSphere	tpEarth, tpMars;	// n`

//--  Planetary rings  --
particle_ring  prSaturn(1.236, 2.329), prUranus(1.487, 2.001);

//--  Textures  --
GLuint	texPlanets[PLANET_NUM];
GLuint	texEarthCloud, texEarthLights, texEarthSpecular, texEarthSpecular2;
GLuint	texEarthDetail[6];	//****
GLuint	texSaturnRing, texUranusRing;

char TopoZoomMode = 0;




//--------------------
//  Register objects
//--------------------
void
register_object_planets()
{
	mtk_object  obj;
	for (int i=0; i<9; i++) {	//******//

		//--  name
		bilingual  name = get_planet_name(i);

		//--  register planets
		obj.init(OBJ_TYPE_PLANET, i, 0.0f, true);
		int obj_idx = register_object(obj, name.text_E);

		//--  register labels
		float R   = 1.4 * _1km_pc * get_planet_Re(MODE_3D, i);
		float thc = 5 * D2R;

		int color_no;
		float name_size;
		if (i==8) {
			color_no  = MTK_COLOR_NAME_DWARF_PLANET;
			name_size = get_dwarf_planet_name_size();
		} else {
			color_no  = MTK_COLOR_NAME_OBJECT;
			name_size = get_object_name_size();
		}

		char key[256];
		sprintf(key, "%sLabel", name.text_E);
		add_label( name.text_J, name.text_E, obj_idx, color_no, 1.0f, name_size, R, thc, true, key );
	}
}




//------------------
//  Landing radius
//------------------
double
get_landing_radius_planet(int idx)
{
	double R = get_planet_Re(MODE_3D, idx) * _1km_pc;

	//---  n`̏ꍇ
	if ( (idx==2 || idx==3)) {
		vector2d  lon_lat = get_landing_pos();
		if (idx==2 && tpEarth.is_loaded()) {
			R += tpEarth.get_elv_fac() * tpEarth.get_elevation(lon_lat.x, lon_lat.y) * _1km_pc / 1000.0;
		} else if (idx==3 && tpMars.is_loaded()) {
			R += tpMars.get_elv_fac() * tpMars.get_elevation(lon_lat.x, lon_lat.y) * _1km_pc / 1000.0;
		}
	}

	return  R;
}



//----------------
//  Prepare draw
//----------------
bool
prepare_draw_planet(const mtk_draw_struct& ds, int idx, double* pRb, double* pRh)
{
	//---  Visible
	bool flg;
	if (idx==8)	flg = test_disp_switch( DISPSW_TNO,		DISPM_DISP);	//--  Pluto
	else		flg = test_disp_switch( DISPSW_PLANET,	DISPM_DISP);
	if (!flg)  return false;

	if (ds.scale > 500 * get_planet_a(idx) * _1AU_pc)  return false;


	//---  Radius
	*pRb = *pRh = get_planet_Re(ds.tyMode, idx) * _1km_pc;
	
	if (hasPlanetRing(idx)) {
		double R1, R2;
		get_planet_ring_radius(idx, &R1, &R2);
		*pRb *= R2;
	}

	return  true;
}



//---------
//  Scale
//---------
double
get_scale_planet(int idx)
{
	return  get_planet_Re(MODE_3D, idx) * _1km_pc * 30.0;
}



//---------
//  Label
//---------
bool
is_planet_label_visible(int idx)
{
	int dispsw_idx;

	if (idx == 8)	dispsw_idx = DISPSW_TNO;
	else			dispsw_idx = DISPSW_PLANET;

	if ( !test_disp_switch(dispsw_idx, DISPM_DISP) || !test_disp_switch(dispsw_idx, DISPM_NAME) )  return  false;

	return  true;
}







//-------------
//  Topo Zoom
//-------------
void
set_topo_zoom_mode(char mode)
{
	TopoZoomMode = mode;
	const float elv_fac = TopoZoomRate[TopoZoomMode];
	tpEarth.set_elv_fac(elv_fac);
	tpMoon.set_elv_fac(elv_fac);
	tpMars.set_elv_fac(elv_fac);
}

char
get_topo_zoom_mode()
{
	return TopoZoomMode;
}




//-----------------
//  Load Textures
//-----------------
void
set_earth_texture0(TopoSphere& tp, GLuint texname, int px1, int py1, int lv, int pw, int ph, int x1, int y1, int tex_w, int tex_h)
{
	const int img_w = 8192;
	const int img_h = 4096;

	float x1p, y1p;
	x1p = float(px1) / 180.0 * img_w;
	y1p = float(py1) /  90.0 * img_h;
	const float u1 = (x1p-x1)/tex_w;
	const float v1 = 1.0 - (y1p-y1)/tex_h;
	const float du =  float(img_w/180.0)/tex_w;
	const float dv = -float(img_h/90.0) /tex_h;

	float v = v1;
	for (int iy=0; iy<ph; iy++) {
		float u = u1;
		for (int ix=0; ix<pw; ix++) {
			patch_texture_info  pti;
			pti.texname = texname;
			pti.u1 = u;
			pti.v1 = v;
			pti.u2 = u + du;
			pti.v2 = v + dv;
			tp.set_patch_texture_info(px1+ix, py1+iy, lv, pti);
			u += du;
		}
		v += dv;
	}

}

void
set_earth_texture()
{
	TopoSphere& tp = tpEarth;
	if (!tp.is_loaded())  return;
	
	tp.set_default_texname( texPlanets[2] );
	set_earth_texture0(tp, texEarthDetail[0],  148, 17, 4, 22,22, 6735,  773, 1024,1024);		// {
	set_earth_texture0(tp, texEarthDetail[1],   9, 33, 4, 10,5,   409, 1501,  256, 128);		// nC
	set_earth_texture0(tp, texEarthDetail[2],   53, 51, 4, 5,10,  2412, 2321,  256, 512);		// `AALMA
}



void
load_planets_textures()
{
	//-- Planets
	for (int i=0; i<PLANET_NUM; ++i) {
		texPlanets[i] = LoadTexture( szPlanetTexFileName[i], TexShrinkRate, false );
	}

	//-- Planetary Ring
	SetSplashText("fÕeNX`Ǎ...", "Loading textures of planetary rings...");
	texSaturnRing = LoadTexture("saturn_rings.png", TexShrinkRate, false);
	texUranusRing = LoadTexture("uranus_rings.png", TexShrinkRate, false);

	//-- Earth Detail
	SetSplashText("n̏ڍ׃eNX`Ǎ...", "Loading detailed textures of Earth...");
	for (int i=0; i<EARTH_DETAIL_TEXTURE_NUM; i++) {
		texEarthDetail[i] = LoadTexture(szEarthDetailTexFileName[i], TexShrinkRate, false);
	}
	set_earth_texture();

	//-- Earth cloud
	SetSplashText("n̉_̃eNX`Ǎ...", "Loading texture of Earth's cloud...");
	texEarthCloud = LoadTexture("earth_cloud.jpg", TexShrinkRate, false);

	//-- Earth lights
	SetSplashText("n̖ĩeNX`Ǎ...", "Loading texture of Earth's lights...");
	texEarthLights = LoadTexture("earth_lights.jpg", TexShrinkRate, false);

	//-- Earth specular
	SetSplashText("ñXyL[}bvǍ...", "Loading texture of Earth's specular map...");
	texEarthSpecular = LoadTexture("earth_specular_with_cloud.jpg", TexShrinkRate, false);

	//-- Earth specular
	SetSplashText("ñXyL[}bvǍ...", "Loading texture of Earth's specular map...");
	texEarthSpecular2 = LoadTexture("earth_specular_without_cloud.jpg", TexShrinkRate, false);
}




//------------------
//  Load Topograph
//------------------
void
load_earth_topograph(directory& dir, const char* filename)
{
	tpEarth.load( dir.get_path(filename) );
	tpEarth.set_elv_fac(TopoZoomRate[TopoZoomMode]);
}

void
load_mars_topograph(directory& dir, const char* filename)
{
	tpMars.load( dir.get_path(filename) );
	tpMars.set_elv_fac(TopoZoomRate[TopoZoomMode]);
	tpMars.shift_texture_coord(180.0f);
}



// ]̌ɍ킹
void
set_planet_orientation(int idx)
{
	set_object_orientation( get_planet_pl_to_ec_matrix(idx) );
}



void
set_planet_light(int idx)
{
	set_light_from_sun( get_planet_position(idx) );
}


//-----------------
// ̓V̂`
//-----------------
void
draw_cloud_sphere(float Rx, float Ry, float Rz)
{
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glDisable(GL_CULL_FACE);
	glPushMatrix();
		glScalef(Rx, Ry, Rz);
		gl_draw_LOD_sphere(true);
	glPopMatrix();
	glEnable(GL_CULL_FACE);
}


//-------------------
//  f̎]`
//-------------------
void
draw_planet_axis(const float Rz)
{
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glLineWidth(1.5);
	glColor3f(0.8, 0.6, 0.2);
	glBegin(GL_LINES);
		glVertex3f(0,0, Rz*1.3);
		glVertex3f(0,0,-Rz*1.3);
	glEnd();
}


//-------
//  n
//-------
vector2d	get_landing_pos();

void
draw_Earth(const mtk_draw_struct& ds, bool bPointer, bool bWire)
{
	const frustum frs = ds.get_frustum();
	const char lang = ds.lang;
	const vector3d  r0 = get_planet_position(PLNT_EARTH);




	glLineWidth(1.0f);
	glColor3f(0,1,0);

	if (tpEarth.is_loaded()) {
		tpEarth.draw(frs, bWire, bDrawLowRes);
	} else {
		glPushMatrix();
			glEnable(GL_LIGHTING);
			glEnable(GL_NORMALIZE);

			glBindTexture(GL_TEXTURE_2D, texPlanets[2]);
			gl_draw_LOD_sphere();
		glPopMatrix();
	}

	const float spc = EarthSpecular;
	float mat0[] = {0,0,0, 1};
	float mat_s[] = {spc, spc, spc, 1};
	glPushAttrib(GL_LIGHTING_BIT | GL_ENABLE_BIT);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_LIGHTING);
		glEnable(GL_NORMALIZE);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);
		glMaterialfv( GL_FRONT_AND_BACK , GL_AMBIENT,    mat0 );
		
		// i
		float difi = 1.0;
		if ( !test_disp_switch(DISPSW_PLANET, DISPM_CITY_LIGHTS) ) {
			const float difi0 = 0.3;
			const double R  = norm(ds.camera.get_pos() - r0) * _1pc_meter;
			const double Re = get_planet_Re(ds.tyMode, 2) * 1000;
			const double h = R - Re;
			difi = difi0 * (1 - h/(1.7*Re));
			if (difi > 1.0) difi = 1.0;
			if (difi < 0.0) difi = 0.0;
		} else {
			float cosphi = dot_product( normalize(r0), normalize(ds.camera.get_pos() - r0) );
			if (cosphi <= 0.0)  difi = 0;
			else				difi *= cosphi;
		}

		if (difi > 0.0) {
			float mat_d[4];
			for (int i=0; i<3; i++) {
				mat_d[i] = difi;
			}
			mat_d[3] = 1.0;

			glMaterialfv( GL_FRONT_AND_BACK , GL_DIFFUSE,    mat_d );
			glBindTexture(GL_TEXTURE_2D, texEarthLights);
			gl_draw_LOD_sphere(false, true);
		}
		
		// Cʂ̋ʔ
		RGBf hl_col;
		{
			const float zrt = get_zoom_rate().zmhPlanet;
			const float H  = 8000 * zrt;		// XP[nCg(H=8km)
			const float n0 = 2.69e25 / zrt;		// number density
			RGBf F = calc_intensity_of_highlight(ds.camera.get_pos(), r0, get_sun_position(), get_planet_Re(ds.tyMode, 2)*1000, 10, H, n0, 1, 30);
			F = spc * F;

			float Fmax = max(F.r, max(F.g, F.b));
			if (Fmax > 1.0) {
				F = correct_RGB(F/Fmax) * Fmax;
			} else {
				F = correct_RGB(F);
			}

			hl_col = F;
		}
		glPushMatrix();
			if (test_disp_switch(DISPSW_PLANET, DISPM_CLOUD)) {
				glBindTexture(GL_TEXTURE_2D, texEarthSpecular);		// _
			}
			else {
				glBindTexture(GL_TEXTURE_2D, texEarthSpecular2);	// _Ȃ
			}

			draw_planet_highlight(ds, 2, hl_col);
		glPopMatrix();

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glPopAttrib();

	// C̕`
	if (AtmosphereMode != ATMOSPHERE_NONE) {
		const float I0 = AtmosphereIntensity;
		const float zrt = get_zoom_rate().zmhPlanet;
		const float H  = 8000 * zrt;		// XP[nCg(H=8km)G
		const float n0 = 2.69e25 / zrt;		// number density
		draw_atmosphere(get_planet_Re(ds.tyMode, 2) * 1000.0, H, 6.0, n0, I0, (AtmosphereMode == ATMOSPHERE_MULTIPLE));
	}

	// _̕\
	if ( test_disp_switch( DISPSW_PLANET, DISPM_CLOUD ) ) {
		glBindTexture(GL_TEXTURE_2D, texEarthCloud);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);
		glDepthMask(GL_FALSE);
		const double h = 10.0/6378.1;	// _̔a̘faƂ̔
		float Rc = 1.0 + h * TopoZoomRate[TopoZoomMode];

		draw_celestial_object(Rc, Rc, Rc);

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthMask(GL_TRUE);
	}

	// H̉e
	if (swEclipse & (ECLIPSE_DISP | ECLIPSE_BORDER)) {
		draw_planet_eclipse(ds, PLNT_EARTH, SAT_MOON);
	}

	// n
	if ( test_disp_switch(DISPSW_PLANET, DISPM_NAME)  &&  ds.tySubMode != SUBMODE_3DCHART) {
		PlaceName& pn = get_place_name(PLACE_OBJ_EARTH);
		bool bHorizontalMode = (get_landing_mode() == LANDING_MODE_HORIZONTAL);
		pn.draw_on_TopoSphere(tpEarth, get_screen_w(), get_screen_z(), get_font(), (lang == LANG_ENG), ds.ltf, bHorizontalMode);
	}

	// |C^
	if (bPointer) {
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		vector2d lon_lat = get_landing_pos();
		draw_pointer_on_TopoSphere(tpEarth, get_screen_w(), get_screen_z(), get_font(), lon_lat.x, lon_lat.y, ds.ltf);
	}

	glEnable( GL_DEPTH_TEST );
	glDisable(GL_NORMALIZE);
}

void
draw_Jupiter_shadow(const mtk_draw_struct& ds, const double& jd)
{
	if (!(swEclipse & (ECLIPSE_DISP | ECLIPSE_BORDER)))  return;

	int sat_num = 4;
	const int sat_idx[] = { SAT_IO, SAT_EUROPA, SAT_GANYMEDE, SAT_CALLISTO };
	for (int i=0; i<sat_num; i++) {
		int idx = sat_idx[i];
		draw_planet_eclipse(ds, PLNT_JUPITER, idx);
	}
}


//------------------------------
//  f
//   ʒuƎ]킹ĂĂ
//------------------------------
void
draw_planet(const mtk_draw_struct& ds, int idx, bool bAxis, bool bFar)
{
	const mtk_object&  tgt = get_target();
	bool bTarget = (tgt.get_type() == OBJ_TYPE_PLANET && tgt.get_index() == idx);

	const char lang = ds.lang;

	const float screen_w = get_screen_w();
	const float screen_z = get_screen_z();
	const float Rz = 1.0 / (1.0f + get_planet_f(idx));



	//---  ]̕`  ---
	glDisable(GL_BLEND);
	if (bAxis) draw_planet_axis(Rz);



	//---  f{̂̕`  ---
	const frustum frs = ds.get_frustum();
	bool bWire = false;	//


	glEnable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);
	glBindTexture(GL_TEXTURE_2D, texPlanets[idx]);


	glPushMatrix();
	glRotatef(180, 0, 0, 1);
	if (ds.tyMode==MODE_3D && (idx==2 || idx==3)) {

		//--- n ---
		if (idx==2) {
			draw_Earth(ds, bSurfacePointer && bTarget && ds.tySubMode != SUBMODE_3DCHART, false);
		}

		//--- ΐ ---
		if (idx==3) {
			glLineWidth(1.0f);
			glColor3f(0,1,0);
			tpMars.draw(frs, bWire, bDrawLowRes);

			if ( test_disp_switch( DISPSW_PLANET, DISPM_NAME )  &&  ds.tySubMode != SUBMODE_3DCHART) {
				PlaceName& pn = get_place_name(PLACE_OBJ_MARS);
				bool bHorizontalMode = (get_landing_mode() == LANDING_MODE_HORIZONTAL);
				pn.draw_on_TopoSphere(tpMars, screen_w, screen_z, get_font(), (lang==LANG_ENG), ds.ltf, bHorizontalMode);
			}

			if (bSurfacePointer && bTarget && ds.tySubMode != SUBMODE_3DCHART) {
				glDisable(GL_LIGHTING);
				glDisable(GL_TEXTURE_2D);
				glEnable(GL_BLEND);
				vector2d lon_lat = get_landing_pos();
				draw_pointer_on_TopoSphere(tpMars, screen_w, screen_z, get_font(), lon_lat.x, lon_lat.y, ds.ltf);
			}

			// C
			if (AtmosphereMode != ATMOSPHERE_NONE) {
				const float I0 = AtmosphereIntensity; //0.3;
				const float  H = 11000;			// XP[nCg(H=11km)
				const float n0 = 4e23;		// number density
				const double h = norm( ds.camera.get_pos() - ds.tgt_pos )*_1pc_km*1000;
				draw_atmosphere(get_planet_Re(ds.tyMode, 3) * 1000.0, H, 8.0, n0, I0, false);
			}
			glDisable(GL_NORMALIZE);
		}
	} else {

		if (idx == PLNT_SATURN) {
			vector3d lv;
			if (bSideLight) {
				lv = get_side_light_ventor(ds);
			}
			else {
				lv = normalize(get_sun_position() - get_planet_position(idx));
			}
			draw_ring_shaded_sphere(idx, lv, prSaturn.get_R1(), prSaturn.get_R2(), texPlanets[PLNT_SATURN], texSaturnRing, 1.0f);
		}
		else if (idx == PLNT_URANUS) {
			vector3d lv;
			if (bSideLight) {
				lv = get_side_light_ventor(ds);
			}
			else {
				lv = normalize(get_sun_position() - get_planet_position(idx));
			}
			draw_ring_shaded_sphere(idx, lv, prUranus.get_R1(), prUranus.get_R2(), texPlanets[PLNT_URANUS], texUranusRing, 0.2f);
		}
		else {
			draw_celestial_object(1, 1, Rz);
		}

		if (idx == PLNT_JUPITER) {
			draw_Jupiter_shadow(ds, ds.JD);
		}


		// n\|C^̕`
		if (bSurfacePointer && bTarget && ds.tySubMode != SUBMODE_3DCHART) {
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glEnable(GL_BLEND);
			glLineWidth(1.0f);
			glColor3f(0,1,0);
			glRotatef(180,0,0,1);
			vector2d lon_lat = get_landing_pos();
			draw_pointer_on_sphere(screen_w, screen_z, get_font(), lon_lat.x, lon_lat.y, ds.ltf);
		}
	}
	glPopMatrix();


	// f̊̕`
	if (hasPlanetRing(idx)) {
		vector3d lv;
		if (bSideLight) {
			lv = get_side_light_ventor(ds);
		}
		else {
			lv = normalize(get_sun_position() - get_planet_position(idx));
		}


		if (idx==PLNT_SATURN)	{	// y̗
			prSaturn.calc_shadow_factor(ds, idx, lv);
			glBindTexture(GL_TEXTURE_2D, texSaturnRing);

			prSaturn.draw(ds, idx, lv);
		}
		if (idx==PLNT_URANUS)	{	// V̗
			prUranus.calc_shadow_factor(ds, idx, lv);
			glBindTexture(GL_TEXTURE_2D, texUranusRing);
			prUranus.draw(ds, idx, lv);
		}
	}

	glDisable(GL_NORMALIZE);
}




//---------------
//  3D `[gp
//---------------
void
draw_planet2(int idx)
{
	glBindTexture(GL_TEXTURE_2D, texPlanets[idx]);
	glRotatef(180,0,0,1);
	const float Rz = 1.0/(1.0f+get_planet_f(idx));
	glDisable(GL_BLEND);

	//---  f{̂̕`  ---
	glEnable(GL_TEXTURE_2D);
	draw_celestial_object(1, 1, Rz);

	// n̉_\
	if ( test_disp_switch(DISPSW_PLANET, DISPM_CLOUD) && idx==2) {
		glBindTexture(GL_TEXTURE_2D, texEarthCloud);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR);
		float Rc = 1.0 + 8 * TopoZoomRate[TopoZoomMode] / get_planet_real_Re(2);
		draw_celestial_object(Rc, Rc, Rc);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}

	if (hasPlanetRing(idx)) {
		glPushMatrix();
			glScalef(1, 1, 1.0/(1.0f+get_planet_f(idx)));
	
			if (idx==5)	{	// y̗
				glBindTexture(GL_TEXTURE_2D, texSaturnRing);
				prSaturn.draw_3Dchart();
			}
			if (idx==6)	{	// V̗
				glBindTexture(GL_TEXTURE_2D, texUranusRing);
				prUranus.draw_3Dchart();
			}
		glPopMatrix();
	}

	glDisable(GL_NORMALIZE);
}




//-----------
//  fO
//-----------
void
register_planets_orbit_curves()
{
	for (int i=0; i<9; i++) {	//***
		int color_no = (i==8) ? MTK_COLOR_ORBIT_DWARF_PLANET : MTK_COLOR_ORBIT_PLANET;
		space_curve  sc( mtk_object(OBJ_TYPE_PLANET, i, 0.0f, false), get_orbit_width(), color_no );
		register_space_curve( sc );
	}
}


void
make_planets_orbit_curve(space_curve& sc, int div_N)
{
	const int idx = sc.get_related_object().get_index();

	sc.clear();
	for (int i=0; i<div_N; i++) {
		double phi = 2.0 * M_PI * double(i) / div_N;
		sc.add_point( calc_planet_orbit_pos(idx, phi) );
	}
}


float
get_planet_label_alpha(const mtk_draw_struct& ds, int idx)
{
	float scl = 1.0f;
	const vector3d  rpos = get_planet_position( idx ) - ds.camera.get_pos();
	float  Rv = scl * calc_visible_radius_planet( idx );
	float  Dth = 2.0f * Rv / rpos.norm();

	return  calc_alpha_by_apparent_diameter( Dth );
}

float
get_planet_orbit_alpha(const mtk_draw_struct& ds, int idx)
{
	const float dscl = 0.035 * _1AU_pc;
	float  alp = calc_alpha_by_disapear_scale_inside( ds.scale, dscl );
	float scl = 1.0f;
	const vector3d  rpos = get_planet_position( idx ) - ds.camera.get_pos();
	float  Rv = scl * calc_visible_radius_planet( idx );
	float  Dth = 2.0f * Rv / rpos.norm();

	return  alp * calc_alpha_by_apparent_diameter( Dth );
}





//------------------
//  Visible radius
//------------------
double
calc_visible_radius_planet(int idx)
{
	return  get_planet_position(idx).norm();
}



//-----------
//  Planets
//-----------
void
draw_a_planet(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const double& scl)
{
	glEnable(GL_LIGHT0);
	float mat_d[] = {1,1,1,1}, mat_a[] = {0.5,0.5,0.5,1};
	mat_a[0] = mat_a[1] = mat_a[2] = get_lighting_ambient();
	glMaterialfv( GL_FRONT_AND_BACK , GL_DIFFUSE,   mat_d );
	glMaterialfv( GL_FRONT_AND_BACK , GL_AMBIENT,   mat_a );


	// Ɩ̐ݒ
	if (!bSideLight)	{ set_planet_light(idx); }
	else				{ set_side_light(ds); }

	const float Re = scl * get_planet_Re(ds.tyMode, idx) * _1km_pc;
	glPushMatrix();
		glTranslated(rpos.x, rpos.y, rpos.z);

		set_planet_orientation(idx);	// ]щ]pZbg

		glEnable(GL_DEPTH_TEST);
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_LIGHTING);

		glScalef(Re, Re, Re);
		draw_planet(ds, idx, test_disp_switch(DISPSW_PLANET, DISPM_AXIS), !ds.is_near_region() );
	glPopMatrix();

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
}





//------------
//  Hit-test
//------------
bool
hit_test_planet(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const vector3d& nv)
{
	matrix3d  M_E2PLNT;
	vector3d  _rpos, _nv;
	M_E2PLNT = get_planet_ec_to_pl_matrix(idx);
	_rpos = M_E2PLNT * rpos;
	_nv = M_E2PLNT * nv;

	// G
	double f;
	f = 1.0 - get_planet_f(idx);

	_rpos.z /= f;
	_nv.z /= f;
	_nv = normalize(_nv);

	// _
	double R, b, c, D, t;
	R = get_planet_Re(MODE_3D, idx) * _1km_pc;
	b = dot_product(_rpos, _nv);
	c = _rpos.norm2() - R*R;
	D = b*b - c;

	if (D <= 0.0)  return false;
	t = -b -sqrt(D);
	if (t < 0.0)  return  false;

	return  true;
}




//--------
//  Info
//--------
void
make_object_info_planet(const mtk_draw_struct& ds, int idx, std::vector<std::string>& text)
{
	bilingual bszName = get_planet_name(idx);

	if (ds.lang == LANG_ENG) {
		text.push_back( bszName.text_E );
		if (idx!=8) { text.push_back( "Type: Planet" ); }
		else {
			text.push_back( "Type: Dwarf Planet" );
			text.push_back( "Minor Planet Number: 134340" );
		}
	} else {
		text.push_back( bszName.text_J );
		if (idx!=8) { text.push_back( "ށFf" ); }
		else {
			text.push_back( "ށFf" );
			text.push_back( "fԍF134340" );
		}
	}
}




//---------------
//  Planetarium
//---------------
void
draw_a_planet_planetarium(const mtk_draw_struct& ds, int idx)
{
	const bilingual bszName = get_planet_name(idx);
	if (ds.tgt_ID == get_object_ID( bszName.text_E ))  return;


	vector3d  ppos = get_planet_position(idx);


	//---  eʒuƃTCY  ---
	vector3d  rpos = ppos - ds.camera.get_pos();
	double  L = rpos.norm();
	// ʒu
	const double R0 = 1.002;
	vector3d pos = R0 * rpos/L;
	// a
	double Re, Rp;
	Re = get_planet_Re(ds.tyMode, idx) * _1km_pc;
	Rp = Re / L;


	// p
	double th = Rp/pos.norm();
	th *= get_planetarium_zooming_factor();

	const double thc1 = 1*ds.rad_per_pixel;
	bool bPoint = (th < thc1);


	//---  Draw  ---
	glDisable(GL_DEPTH_TEST);	//****

	glPushMatrix();
		glTranslatef(pos.x, pos.y, pos.z);
	
		if (!bPoint) {
			//---  TCYŕ`  ---
			glPushMatrix();
				set_light_from_sun( ppos );
				set_planet_orientation(idx);

				glEnable(GL_NORMALIZE);
				glEnable(GL_LIGHTING);
				glScalef(Rp, Rp, Rp);
				draw_planet(ds, idx, false, false);
			glPopMatrix();
		} else {
			//---  _ŕ`  ---
			glDisable(GL_LIGHTING);
			glDisable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);
			glPointSize(1.5);
			glBegin(GL_POINTS);
				glColor3f(1,1,1);
				glVertex3f(0,0,0);
			glEnd();
		}

		// O
		if ( test_disp_switch(DISPSW_PLANET, DISPM_NAME) ) {
			float sz;
			int col_idx;

			if (idx != PLNT_PLUTO) {
				sz = get_object_name_size();
				col_idx = MTK_COLOR_NAME_OBJECT;
			} else {
				sz = get_dwarf_planet_name_size();
				col_idx = MTK_COLOR_NAME_DWARF_PLANET;
			}

			prepare_3Dtext_1(Rp*1.2*ds.plnt_R, 45, sz*ds.ltf, 1.0f, false);
			
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glEnable(GL_BLEND);

			const RGBf col = get_color(col_idx);
			glColor4f(col.r, col.g, col.b, col_star_name.a);
			draw_bilingual(bszName);
		}
	glPopMatrix();
}




void
draw_planets_planetarium(const mtk_draw_struct& ds)
{
	float mat_d[] = {1,1,1,1};
	float mat_a[] = {0.5,0.5,0.5,1};
	mat_a[0] = mat_a[1] = mat_a[2] = get_lighting_ambient();
	glMaterialfv( GL_FRONT_AND_BACK , GL_DIFFUSE,   mat_d );
	glMaterialfv( GL_FRONT_AND_BACK , GL_AMBIENT,   mat_a );

	glEnable(GL_LIGHT0);

	for (int i=0; i<9; i++) {
		draw_a_planet_planetarium(ds, i);
	}
}

